# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by signac.core classes."""


class Error(Exception):
    """Base class used for signac Errors."""

    pass
