# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define signac version."""


__version__ = "1.5.1"

SCHEMA_VERSION = "1"


__all__ = [
    "__version__",
    "SCHEMA_VERSION",
]
