# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Migrate from schema version 0 to version 1.

This migration is a null-migration that serves as a template
for future migrations and testing purposes.
"""


def migrate_v0_to_v1(project):
    """Migrate from schema version 0 to version 1."""
    pass  # nothing to do here, serves purely as an example
